﻿Imports Microsoft.VisualBasic
Imports System
Imports System.IO
Imports System.Collections
Imports System.ComponentModel

''' <summary> Zeigt die Ordnerstruktur des Dateisystems an </summary>
Public Class FolderTreeview : Inherits FileSystemTreeView

    Private WithEvents _NodeFiller As New JobWorker

    Protected Overrides Sub OnHandleCreated(ByVal e As System.EventArgs)
        MyBase.OnHandleCreated(e)
        ' Me.DesignMode funktioniert erst bei vorhandenem Control-Handle
        If Me.DesignMode Then Return
        MyBase.Sorted = True 'Simpel-Sortierung reicht hier
        CheckDrives()
    End Sub

#Region "Public_Members"

    <Browsable(False)> _
    <DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)> _
    Public Property SelectedFolder() As String
        Get
            If MyBase.SelectedNode Is Nothing Then Return ""
            Return MyBase.SelectedNode.FullPath
        End Get
        Set(ByVal NewFolder As String)
            If SelectedFolder = NewFolder Then Return
            If Directory.Exists(NewFolder) Then
                GotoDirectory(NewFolder)
            Else
                MessageBox.Show(Me, _
                    String.Format("'{0}' kann nicht gefunden werden", NewFolder), _
                    My.Application.Info.Title, _
                    MessageBoxButtons.OK, _
                    MessageBoxIcon.Information)
            End If
        End Set
    End Property

    ''' <summary>
    '''  Bringt (nur) die zugreifbaren Laufwerke zur Anzeige 
    '''   (ggfs. Wechselmedium einlegen)
    ''' </summary>
    Public Sub CheckDrives()
        NodeFillJob.FillWithDrives(Me.Nodes)
        For Each Nd As TreeNode In Me.Nodes
            Nd.ImageIndex = NodeType.Drive
            Nd.SelectedImageIndex = NodeType.Drive
        Next
    End Sub

#End Region 'Public_Members

#Region "Overrides, Events"

    Protected Overrides Sub OnBeforeSelect(ByVal e As TreeViewCancelEventArgs)
        ' Bei Selektion eines ungültigen Nodes auf den nächsten gültigen 
        '  Parent wechseln
        Dim Nd As TreeNode = GetValidNode(e.Node)
        If Nd IsNot e.Node Then
            e.Cancel = True
            Me.SelectedNode = Nd
            Return
        End If
        MyBase.OnBeforeSelect(e)
    End Sub

    Protected Overrides Sub OnAfterExpand( _
            ByVal e As System.Windows.Forms.TreeViewEventArgs)

        Dim Nd = e.Node
        ' Abbruch, wenn nicht Nd im Dateisystem existiert
        If Nd IsNot GetValidNode(Nd) Then Return

        ' Da wir keinen FilesystemWatcher einsetzen möchte, synchronisieren wir
        '  beim Öffnen eines Nodes jedesmal neu mit dem Dateisystem.
        _NodeFiller.Add(AddressOf (New NodeFillJob(Nd)).GetSubFolders)
        MyBase.OnAfterExpand(e)
    End Sub

    Protected Overrides Sub WndProc(ByRef m As System.Windows.Forms.Message)
        ' OnPaint wird nicht aufgerufen, daher testen wir in WndProc auf WM_PAINT
        Const WM_PAINT As Integer = &HF
        MyBase.WndProc(m)
        If m.Msg = WM_PAINT Then
            ' Bei jedem Paint werden alle sichtbaren, noch nicht befüllten Nodes
            '  mit dem Dateisystem synchronisiert. Die Belastung des 
            '  Main-Threads ist jedoch minimal, wg. Einsatz von Threading.
            ' Die Gesamt-Entlastung besteht darin, daß auch bei großem 
            '  Bildschirm kaum jemals mehr als 40 Nodes sichtbar sind.
            Using _NodeFiller.Monitor
                Dim Nd = Me.TopNode
                Do Until Nd Is Nothing OrElse Not Nd.IsVisible
                    If Nd.Tag IsNot Nothing Then
                        ' Die zu befüllenden Nodes sind durch 
                        '  "Nd.Tag <> Nothing" markiert
                        _NodeFiller.Add( _
                            AddressOf (New NodeFillJob(Nd)).GetSubFolders)
                    End If
                    Nd = Nd.NextVisibleNode
                Loop
            End Using
        End If
    End Sub

    Private Sub _NodeFiller_AllDone( _
            ByVal sender As Object, _
            ByVal e As JobWorker.EventArg) Handles _NodeFiller.AllDone

        For Each Job As NodeFillJob In e.DoneJobs
            ' Der Nebenthread hat Dateisuchen durchgeführt und Ordner-Listen 
            '  erstellt. Treenodes einhängen muß leider im Hauptthread erfolgen.
            Job.Synchronize()
        Next
    End Sub

#End Region 'Overrides, Events

#Region "Privates"

    Private Function GetValidNode(ByVal Nd As TreeNode) As TreeNode
        If Directory.Exists(Nd.FullPath) Then Return Nd
        Do Until Directory.Exists(Nd.FullPath)
            Nd = Nd.Parent
        Loop
        _NodeFiller.Add(AddressOf (New NodeFillJob(Nd)).GetSubFolders)
        Return Nd
    End Function


    Private Sub Kanonisize(ByRef sFullPath As String)
        ' Ein Gemurkel mit Microsoft.VisualBasic.FileSystem.Dir(), um die 
        '  Groß/Klein-Schreibweise zu ermitteln, die im Dateisystem 
        '  hinterlegt ist.
        Dim Segments = sFullPath.Split(_Separator)
        With New System.Text.StringBuilder(Segments(0))
            For I = 1 To Segments.Length - 1
                .Append(_Separator).Append( _
                    Dir(.ToString & Segments(I), FileAttribute.Directory))
            Next
            sFullPath = .ToString
        End With
    End Sub

    Private Sub GotoDirectory(ByVal sFullPath As String)
        ' Es wird der sFullPath nächste übergeordnete Treenode gesucht, und von 
        '  dem aus ein Ketten-Abschnitt von Nodes erstellt, sodass die 
        '  Gesamt-Kette sFullPath korrekt darstellt. EnsureVisible auf das 
        '  unterste Kettenglied angewandt expandiert alle darüber liegenden 
        '  Nodes und löst damit die Befüllung derselben in OnAfterExpand() aus.
        Kanonisize(sFullPath)
        With Me.Approach(sFullPath)
            Dim Nd = .Node
            If Not .IsFullMatch Then
                For I = .StepsDone To .Segments.Count - 1
                    Nd = Nd.Nodes.Add(.Segments(I))
                    NodeFillJob.Fill(Nd)
                Next
            End If
            Using _NodeFiller.Monitor
                Nd.EnsureVisibleX(StringAlignment.Center, True)
            End Using
            Me.SelectedNode = Nd
        End With
    End Sub

#End Region 'Privates

End Class